// A Java version for the Kaser Interpreter - the general UI functions 
// For some of my own fun!
// Stephen M. Bennett
// 29Jun02 Started

package stephen.bennett.kint;


// We are going to do the UI through Swing

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;
import javax.swing.*;
import javax.imageio.*;


public class UI {
  // boolean to be set if direct drawing it to be employed
  private static boolean directDraw = false;

  // the screen width, height, and size index
  private static int screenWidth;
  private static int screenHeight;
  private static int screenWidthIndex;

  // we maintain our own invalid paint area, for buffered drawing operations
  // null means no invalidated Rectangle
  private static Rectangle invalidPaintArea = null;

  // for trace statements to check out invalid paint area
  // private static Vector invalidRectangles = new Vector();

  // our UI objects
  private static JFrame jFrame;
  private static KPanel kPanel;
  private static javax.swing.Timer timer;
  private static EventQueue awtEventQueue;

  // the current text colour, back colour and mode and font
  private static Color currentTextColour = Color.BLACK;
  private static Color currentBackColour = Color.WHITE;
  private static int currentBackMode = TextState.BACKMODE_TRANSPARENT;
  private static Font currentFont = new Font("Arial", Font.PLAIN, 12);

  // a 1x1 image for setPixel direct to screen
  private static Image image1by1;

  // the time the last (interesting) event happened
  private static long mostRecentEventTime = 0L;

  // the defined fonts - non-null entry means defined, null is a free slot
  // each element is Font, handle is same as index
  private static Vector fonts = new Vector();

  // the defined BMBs - bitmap handle is index into this Vector
  // 0 means the game's display bitmap (assured because it is created first, hopefully :-)
  // null entries are free slots and the handle can be re-used
  // each element is BMB
  private static Vector bmbs = new Vector();

  // the Text states for SYS_BEGTEXT, SYS_ENDTEXT
  // each element is TextState
  private static Vector textStates = new Vector();

  // used to indicate that the next move and size events should be ignored
  // also need to keep track of current location and size
  private static boolean ignoreNextWindowMove = false;
  private static boolean ignoreNextWindowSize = false;
  private static int jFrameX = 0;
  private static int jFrameY = 0;
  private static int jFrameW = 0;
  private static int jFrameH = 0;

  // boolean to say if we should use .PNG files instead, false means .BMP
  private static boolean usePNG = false;

  // some performance info
  private static long loadBitmapTime = 0;
  private static long loadMBitmapTime = 0;

  // cursor stuff
  public static final int CURSOR_NONE   = -1;
  public static final int CURSOR_NORMAL = 0;
  public static final int CURSOR_WAIT   = 1;
  public static final int CURSOR_HELP   = 2;
  private static int currentCursor = CURSOR_NORMAL;
  
  // BMB rotation/flip orient values
  public static final int BMO_NORMAL  = 0; // normal orientation
  public static final int BMO_90      = 1; // rotated counter-clockwise 90
  public static final int BMO_180     = 2; // rotated counter-clockwise 180
  public static final int BMO_270     = 3; // rotated counter-clockwise 270
  public static final int BMO_FNORMAL = 4; // flipped left-right
  public static final int BMO_F90     = 5; // flipped & rotated 90
  public static final int BMO_F180    = 6; // flipped & rotated 180
  public static final int BMO_F270    = 7; // flipped & rotated 270
  


  //***************************************************
  // screen methods
  //***************************************************

  // initialise the screen 
  // called at start up and on ?WM_DISPLAYCHANGE? event
  public static void initScreen() {
     // find screen sizes and convert to index
    Toolkit toolkit = Toolkit.getDefaultToolkit();
    Dimension dimension = toolkit.getScreenSize();
    
    // Print size
    screenWidth = dimension.width;
    screenHeight = dimension.height;
    if (screenWidth < 800)  screenWidthIndex = 0;
    else if (screenWidth < 1024) screenWidthIndex = 1;
    else if (screenWidth < 1152) screenWidthIndex = 2;
    else if (screenWidth < 1280) screenWidthIndex = 3;
    else if (screenWidth < 1600) screenWidthIndex = 4;
    else                         screenWidthIndex = 5;
    Log.fine("screenWidth", screenWidth);
    Log.fine("screenHeight", screenHeight);
    Log.fine("screenWidthIndex", screenWidthIndex);

    // cope if called if we can determin Screen Changed resolution
    //  if( redraw ) {
    //    // call the SCREEN_CHANGE code in the KE module
    //    KEevent(KHDR_SCREEN_CHANGE, 0, 0, 0);
    //  }
  }


  // return the screen width
  public static int getScreenWidth() {
    return screenWidth;
  }

  // return the screen height
  public static int getScreenHeight() {
    return screenHeight;
  }

  // return the screen width index
  public static int getScreenWidthIndex() {
    return screenWidthIndex;
  }




  //*********************************************
  // Window Creation, Location and Size methods
  //*********************************************


  // create the window (KPanel)
  public static void createWindow() {
    Log.fine("createWindow");

    // create a new frame
    jFrame = new JFrame("KInt Loading, Please Wait...");

    // we will handle the close in the Window listener
    jFrame.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

    // create the KPanel and put it in the frame
    kPanel = new KPanel();
    jFrame.getContentPane().add(kPanel);
    
    // need to show the window
    // let's have a default size as well - will be resized very quickly by
    // most (if not all) .KE programs
    jFrame.setSize(300, 300);
    jFrame.show();
    
    // create the BMB to represent this display
    createBMB(300, 300, true);
    
    // we are interested a number of events, so we call call .KE functionality
    // setting after show so that we only get events after the show
    jFrame.addWindowListener(new KWindowListener());
    jFrame.addComponentListener(new KComponentListener());
    jFrame.addKeyListener(new KKeyListener());
    //kPanel.addKeyListener(new KKeyListener());
    kPanel.addMouseListener(new KMouseListener());
    kPanel.addMouseMotionListener(new KMouseMotionListener());

    // we do our own focus travelling, thank you (for TAB char input)
    jFrame.setFocusTraversalKeysEnabled(false);

    // remember our location and size - we need to keep track of it
    jFrameX = getX();
    jFrameY = getY();
    jFrameW = getWidth();
    jFrameH = getHeight();
  }

  // show the window
  public static void showWindow() {
    jFrame.show();
  }

  // return the window's (JFrame's) X
  public static int getX() {
    return jFrame.getX();
  }

  // return the window's (JFrame's) Y
  public static int getY() {
    return jFrame.getY();
  }

  // set the window's (JFrame's) location
  public static void setLocation(int x, int y) {
    // When we are setting this ourselves, the rules are that .KE does NOT
    // get a WINDOW_MOVE event.  So, we remember we have set it and
    // clear the flag when the Java event happens (not sure what to do if
    // the event does not happen for some reason, .KE will not be told about
    // the next window move that the user did).  

    // Actually if jFrame.setLocation is to current location, then Java does
    // not generate the event - so the ignoreNextWindowMove will get out
    // of step - so I am going to remember current location and only
    // set the location on the frame if its for a new location
    if (x == jFrameX && y == jFrameY) {
      Log.fine("setLocation not done as it is to the current location");
      return;
    }

    // this will generate an event, so ignore the next window move
    setIgnoreNextWindowMove(true);

    // set, and remember, the location
    Log.fine("setLocation", x, y);
    jFrame.setLocation(x, y);
    jFrameX = x;
    jFrameY = y;
  }


  // get the ignore next window move boolean
  public static boolean getIgnoreNextWindowMove() {
    return ignoreNextWindowMove;
  }

  // set the ignore next window move boolean
  public static void setIgnoreNextWindowMove(boolean b) {
    Log.fine("setting ignore next window move to", b);
    ignoreNextWindowMove = b;
  }


  // return the window's (JFrame's) Width
  public static int getWidth() {
    return jFrame.getWidth();
  }

  // return the window's (JFrame's) Height
  public static int getHeight() {
    return jFrame.getHeight();
  }

  // set the window's (JFrame's) size
  public static void setSize(int w, int h) {
    // similar comment about .KE not getting WINDOW_SIZE if .KE did the set
    // size (as for setLocation).  We must not call setSize if it is for the
    // same size
    if (w == jFrameW && h == jFrameH) {
      Log.fine("setSize not done as it is to the current size");
      return;
    }

    // this will generate an event, so ignore the next size event
    setIgnoreNextWindowSize(true);

    // set and remember the size
    Log.fine("setSize", w, h);
    jFrame.setSize(w, h);
    jFrameW = w;
    jFrameH = h;

    // need to validate so that KPanel resizes as well
    jFrame.validate();
  }

  // get the ignore next window size boolean
  public static boolean getIgnoreNextWindowSize() {
    return ignoreNextWindowSize;
  }

  // set the ignore next window size boolean
  public static void setIgnoreNextWindowSize(boolean b) {
    Log.fine("setting ignore next window size to", b);
    ignoreNextWindowSize = b;
  }

  // return the window's (JFrame's) extended state
  public static int getExtendedState() {
    return jFrame.getExtendedState();
  }

  // set the window's (JFrame's) extended state
  public static void setExtendedState(int v) {
    jFrame.setExtendedState(v);
  }

  // set the window (JFrame) text (title)
  public static void setWindowText(String text) {
    Log.fine("setWindowText", text);
    jFrame.setTitle(text);
  }


  // destroy the JFrame
  public static void destroyWindow() {
    Log.fine("destroyWindow");
    jFrame.dispose();
  }


  //************************************************
  // window view methods
  //************************************************

  // get the window's (KPanel's) view size width
  public static int getViewWidth() {
    int w = kPanel.getWidth();
    Log.fine("getViewWidth returing", w);
    return w;
  }

  // get the window's (KPanel's) view size height
  public static int getViewHeight() {
    int h = kPanel.getHeight();
    Log.fine("getViewHeight returning", h);
    return h;
  }


  //*****************************************************
  // Graphic Operations
  //****************************************************


  // add the following rect to the invalid paint area
  private static void invalidateRect(int x, int y, int w, int h) {
    //System.out.println("invalid rect (x,y)=("+x+","+y+")");
    //System.out.println("invalid rect (w,h)=("+w+","+h+")");
    Log.fine("invalidateRect x y", x, y);
    Log.fine("invalidateRect w h", w, h);

    // for tracing
    // invalidRectangles.add(new Rectangle(x, y, w, h));

    if (invalidPaintArea == null) {
      Log.fine("invalidateRect New invalid area being created");
      invalidPaintArea = new Rectangle(x, y, w, h);
    }
    else {
      Log.fine("invalidateRect Extending existing invalid area");
      invalidPaintArea.add(x, y);
      invalidPaintArea.add(x+w, y+h);
    }

    // printInvalidPaintArea("invalidPaintArea is now");
  }


  // just for tracing
  private static void printInvalidPaintArea(String msg) {
    System.out.println(msg+" ("+invalidPaintArea.x+","+invalidPaintArea.y+") w="+invalidPaintArea.width+", h="+invalidPaintArea.height); 
  }


  // paint the invalid area now
  // this results in there no longer being a invalid paint area
  public static void paintInvalidArea() {
    if (invalidPaintArea == null) {
      Log.fine("paintInvalidArea nothing to do");
    }
    else {
      // for tracing
      // System.out.println("paintInvalidArea constructed from:");
      // for(int i=0; i<invalidRectangles.size(); i++) {
      //   Rectangle r = (Rectangle)invalidRectangles.elementAt(i);
      //   System.out.println("  "+r);
      // }
      // System.out.println("Full area is "+invalidPaintArea);
      // invalidRectangles = new Vector();

      // get the off-screen buffered image
      Image image = kPanel.getImage();
      if (image == null) {
        Log.errorExit("paintInvalidArea offscreen image is null");
      }

      // get the graphics for the screen
      Graphics g = kPanel.getGraphics();
      if (g == null) {
        // this seems to be when we are exiting?
        Log.fine("paintInvalidArea graphics is null");
        return;
      }

      // printInvalidPaintArea("Painting invalid area");

      // compute the args, same for src and dest
      int x1 = invalidPaintArea.x;
      int y1 = invalidPaintArea.y;
      int x2 = x1 + invalidPaintArea.width;
      int y2 = y1 + invalidPaintArea.height;

      // draw the part of the image that is invalid
      g.drawImage(image, x1, y1, x2, y2, x1, y1, x2, y2, null);

      // area painted okay
      invalidPaintArea = null;
    }
  }



  // switch to direct drawing
  public static void useDirectDraw() {
    directDraw = true;
  }


  // append .PNG or .BMP to give String buffer
  public static void appendGraphicsFileType(StringBuffer sb) {
    if (usePNG) {
      sb.append(".PNG");
    }
    else {
      sb.append(".BMP");
    }
  }

  // report if PNG is being used
  public static boolean usingPNG() {
    return usePNG;
  }


  // switch to using PNG files
  public static void usePNGFiles() {
    usePNG = true;
  }


  // set the cursor value
  public static void setCursor(int c) {
    currentCursor = c;
  }

  // show the current cursor
  public static void showCursor() {
    Log.fine("showCursor", currentCursor);

    // none is now normal
    if (currentCursor == CURSOR_NONE) {
      currentCursor = CURSOR_NORMAL;
    }

    // show this cursor
    if (currentCursor == CURSOR_HELP) {
      kPanel.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
    }
    else if (currentCursor == CURSOR_WAIT) {
      kPanel.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    }
    else {
      kPanel.setCursor(Cursor.getDefaultCursor());
    }
  }


  // create a bitmap of the given size, returning bitmap handle, -1 for failure
  public static int createBMB(int w, int h, boolean sysBM) {
    Log.fine("createBMB", w, h);

    // need the Image to hold the Bitmaps
    Image image = null;
    if (sysBM) {
      // this is KPanel's image
      image = kPanel.getImage();
    }
    else {
      // use KPanel to create the actual Image to hold the Bitmap
      image = kPanel.createImage(w, h);
    }
    
    if (image == null) {
      Log.severe("Could not create image in kPanel width x height:", w, h);
      return -1;
    }
    
    // create the BMB
    BMB bmb = new BMB(w, h, image);

    // store and return bitmap handle 
    // *** getting game's display to be handle 0 relies on the fact this will be called first
    return storeBitmap(bmb);
  }


  // delete the given bitmap from its handle
  // for us it is just remove the bitmap from the remembered list
  // not allowed to do for handle 0 (the screen)
  public static void deleteBitmap(int bmbHandle) {
    Log.fine("deleteBitmap", bmbHandle);

    if (bmbHandle == 0) {
      // handle 0 is screen, that cannot be deleted
      Log.info("deleteBitmap called with handle 0");
      return;
    }

    Utils.removeObjectByHandle(bmbHandle, bmbs, false, "BMBs");
  }





  // load the bitmap file, depending on .BMP or .PNG setting
  // does transparency if trans given for colour given
  private static BufferedImage loadBitmapFile(String filename, boolean trans,
                                              int transX, int transY) {
    try {
      if (usePNG) {
        // Get the file first, then do transparency if required
        BufferedImage image = ImageIO.read(new File(filename));

        // process transparency if requested
        if (trans) {
          // for PNG there is not alpha channel, so create another BufferedImage
          // and copy all bits across, setting alpha to 0 for any that have the
          // same colour as pixel named

          // get the RGB colour, with no alpha
          int transparentRGB = image.getRGB(transX, transY) & 0xFFFFFF;
          Log.fine("transparent pixel RGB value", transparentRGB);

          // create an RGB image with ALPHA values of the same size
          int w = image.getWidth(null);
          int h = image.getHeight(null);
          BufferedImage result = new BufferedImage(w, h, 
                                                   BufferedImage.TYPE_INT_ARGB);

          // go through all pixels, setting alpha to 0 for any that match our rgb
          for(int x=0; x<w; x++) {
            for(int y=0; y<h; y++) {
              int rgb = image.getRGB(x, y);
              if ((rgb & 0xFFFFFF) == transparentRGB) {
                // we want it transparent
                result.setRGB(x, y, transparentRGB);
              }
              else {
                // use original
                result.setRGB(x, y, rgb);
              }
            }
          }
          
          // this is the new image
          image = result;
        }

        // return the image
        return image;
      }
      else {
        if (trans) {
          return WindowsBMP.read(new File(filename), transX, transY);
        }
        else {
          return WindowsBMP.read(new File(filename));
        }
      }
    }
    catch (IOException e) {
      Log.severe("Exception in reading image \""+filename+"\" : "+e);
      return null;
    }
  }


  // load a bitmap - returning bitmap handle
  public static int loadBitmap(String filename) {
    // for performance
    long t1 = System.currentTimeMillis();

    // get the image from the file (no transparent pixels)
    BufferedImage image = loadBitmapFile(filename, false, -1, -1);

    // for performance
    long t2 = System.currentTimeMillis();
    loadBitmapTime += (t2-t1);

    if (image == null) {
      Log.severe("Null image returned while loading bitmap", filename);
      return 0;
    }
    // Log.errorExit("image is "+image);

    // store and return bitmap handle
    return storeBitmap(image);
  }


  // load a MBitmap - returning bitmap handle
  // x and y are transparent pixel co-ords
  public static int loadMBitmap(String filename, int transX, int transY) {
    Log.fine("loadMBitmap transparent x y", transX, transY);

    // for performance
    long t1 = System.currentTimeMillis();

    // get the image from the file
    BufferedImage image = loadBitmapFile(filename, true, transX, transY);

    // for performance
    long t2 = System.currentTimeMillis();
    loadMBitmapTime += (t2-t1);

    if (image == null) {
      Log.severe("Null image returned while loading bitmap", filename);
      return 0;
    }

    // store and return bitmap handle
    return storeBitmap(image);
  }


  // get the pixel at the given co-ords for the given bitmap handle
  public static int getPixel(int bitmapHandle, int x, int y) {
    Log.fine("getPixel x y", x, y);

    // get the image
    BufferedImage image = (BufferedImage)getImage(bitmapHandle);

    int bgr = swapRedAndBlue(image.getRGB(x, y));
    Log.fine("getPixel result", bgr);
    
    return bgr;
  }


  // set the pixel at the given co-ords for the given bitmap handle to the
  // given BGR value - if (x,y) is outside of image then do nothing
  // (have to cope with this as .KE does do this)
  public static void setPixel(int bitmapHandle, int x, int y, int bgr) {
    Log.fine("setPixel x y", x, y);
    Log.fine("setPixel bgr", bgr);

    // talk in Java RGB terms
    int rgb = swapRedAndBlue(bgr);

    // get the image
    BufferedImage image = (BufferedImage)getImage(bitmapHandle);

    // check that x,y are within the image
    if (x < 0 || x >= image.getWidth()) {
      Log.fine("ignoring setPixel as x outside of image");
      return;
    }
    if (y < 0 || y >= image.getHeight()) {
      Log.fine("ignoring setPixel as y outside of image");
      return;
    }

    // set the pixel
    image.setRGB(x, y, rgb);

    // if screen involved, draw direct or repaint
    if (bitmapHandle == 0) {
      if (directDraw) {
        // if we are drawing direct to screen, then need to do special as
        // no way to set direct pixel with JPanel.  I hold my own image
        // of size 1x1, set pixel on that, then draw direct to screen

        // create the 1x1 image if we don't have it yet
        if (image1by1 == null) {
          image1by1 = kPanel.createImage(1, 1);
          if (image1by1 == null) {
            Log.errorExit("Cannot create 1x1 image for setPixel");
          }
        }

        // set the pixel
        ((BufferedImage)image1by1).setRGB(0, 0, rgb);
        
        // draw it onto the real screen
        drawBitmap1(image1by1, x, y, kPanel.getGraphics());
      }
      else {
        // invalidate this pixel rectangle, so flush draws this to screen
        invalidateRect(x, y, x+1, y+1);

        // repaint();
      }
    }
  }


  // draw one bitmap on another
  public static void drawBitmap(int bitmapHandleDst, int bitmapHandleSrc, 
                                int xDst, int yDst, 
                                int xSrc, int ySrc, int wSrc, int hSrc) {
    Log.fine("drawBitmap xDst yDst", xDst, yDst);
    Log.fine("drawBitmap xSrc ySrc", xSrc, ySrc);
    Log.fine("drawBitmap wSrc hSrc", wSrc, hSrc);

    // get the source image
    BufferedImage srcImage = (BufferedImage)getImage(bitmapHandleSrc);

    // we only want part of it, first need to make sure we are not asking
    // for too much.  Java throws exception in getSubimage if out of bounds
    int srcImageWidth = srcImage.getWidth(null);
    if (srcImageWidth < 0) {
      // cannot cope
      Log.errorExit("Cannot find width of image for handle", bitmapHandleSrc);
    }

    int srcImageHeight = srcImage.getHeight(null);
    if (srcImageHeight < 0) {
      // cannot cope
      Log.errorExit("Cannot find height of image for handle", bitmapHandleSrc);
    }

    // reduce width asking for, if it means we would go beyond image width
    int srcWidth = xSrc+wSrc;
    if (srcWidth > srcImageWidth) {
      Log.fine("Having to reduced requested width", srcWidth, srcImageWidth);
      wSrc = wSrc - (srcWidth - srcImageWidth);
    }

    // reduce height asking for, if it means we would go beyond image height
    int srcHeight = ySrc+hSrc;
    if (srcHeight > srcImageHeight) {
      Log.fine("Having to reduced requested height", srcHeight, srcImageHeight);
      hSrc = hSrc - (srcHeight - srcImageHeight);
    }

    // get sub image (all the image if that is being requested)
    Image image = srcImage;
    if (xSrc != 0 || ySrc != 0 ||
        wSrc != srcImageWidth || hSrc != srcImageHeight) {
      image = srcImage.getSubimage(xSrc, ySrc, wSrc, hSrc);
    }

    // draw the image 
    drawBitmap1(image, xDst, yDst, getImage(bitmapHandleDst).getGraphics());

    // if screen involved, draw direct or repaint
    if (bitmapHandleDst == 0) {
      if (directDraw) {
        drawBitmap1(image, xDst, yDst, kPanel.getGraphics());
      }
      else {
        // invalidate the rectangle, so flush draws this to screen
        invalidateRect(xDst, yDst, wSrc, hSrc);

        // repaint();
      }
    }

  }


  // the common method for drawBitmap (for possible double operation)
  private static void drawBitmap1(Image image, int xDst, int yDst, 
                                  Graphics g) {
    // we check completion (though I don't handle if we fail)
    if (!g.drawImage(image, xDst, yDst, null)) {
      Log.errorExit("drawBitmap1: drawImage did not complete, cannot progress");
    }
  }


  // draw masked bitmap on another bitmap
  // for us, the image already knows which bits are transparent
  public static void drawMBitmap(int bitmapHandleDst, int bitmapHandleSrc, 
                                int xDst, int yDst, 
                                int xSrc, int ySrc, int wSrc, int hSrc) {
    drawBitmap(bitmapHandleDst, bitmapHandleSrc, xDst, yDst, 
               xSrc, ySrc, wSrc, hSrc);
  }


  
  // draw one bitmap on another, stretching it
  public static void stretchBitmap(int bitmapHandleDst, int bitmapHandleSrc, 
                                   int xDst, int yDst, int wDst, int hDst,
                                   int xSrc, int ySrc, int wSrc, int hSrc) {
    Log.fine("stretchBitmap xDst yDst", xDst, yDst);
    Log.fine("stretchBitmap wSrc hSrc", wSrc, hSrc);
    Log.fine("stretchBitmap xSrc ySrc", xSrc, ySrc);
    Log.fine("stretchBitmap wSrc hSrc", wSrc, hSrc);

    // get the source image
    Image srcImage = getImage(bitmapHandleSrc);

    // compute the necessary values
    int xDst2 = xDst+wDst;
    int yDst2 = yDst+hDst;
    int xSrc2 = xSrc+wSrc;
    int ySrc2 = ySrc+hSrc;

    // draw the image, scaling as required
    stretchBitmap1(srcImage, 
                   xDst, yDst, xDst2, yDst2,
                   xSrc, ySrc, xSrc2, ySrc2,
                   getImage(bitmapHandleDst).getGraphics());

    // if screen involved, draw direct or repaint
    if (bitmapHandleDst == 0) {
      if (directDraw) {
        stretchBitmap1(srcImage, 
                       xDst, yDst, xDst2, yDst2,
                       xSrc, ySrc, xSrc2, ySrc2,
                       kPanel.getGraphics());
      }
      else {
        // invalidate the rectangle, so flush draws this to screen
        invalidateRect(xDst, yDst, wDst, hDst);

        // repaint();
      }
    }
  }


  // the common method for stretchBitmap (for possible double operation)
  private static void stretchBitmap1(Image srcImage, 
                                     int xDst, int yDst, int xDst2, int yDst2,
                                     int xSrc, int ySrc, int xSrc2, int ySrc2, 
                                     Graphics g) {
    if (!g.drawImage(srcImage, xDst, yDst, xDst2, yDst2,
                     xSrc, ySrc, xSrc2, ySrc2, null)) {
      Log.errorExit("stretchBitmap1: drawImage did not complete, cannot progress");
    }
  }

 

  
  // draw an edged rectangle on given bitmap handle
  // for Java do as outline rectangle, and filled rectangle
  public static void rectEdged(int bitmapHandle, int fillColour, int frameColour,
                               short left, short top,
                               short right, short bottom) {
    Log.fine("rectEdged fillColour frameColour", fillColour, frameColour);
    Log.fine("rectEdged left top", left, top);
    Log.fine("rectEdged right bottom", right, bottom);

    // compute necessary values
    int width = right-left;
    int height =bottom-top;
    Color fillc = getColorFromBGR(fillColour);
    Color framec = getColorFromBGR(frameColour);

    // draw a filled rect then edge it
    rectEdged1(fillc, framec, left, top, width, height, 
               getImage(bitmapHandle).getGraphics());

    // if screen involved, draw direct or repaint
    if (bitmapHandle == 0) {
      if (directDraw) {
        rectEdged1(fillc, framec, left, top, width, height, 
                   kPanel.getGraphics());
      }
      else {
        // invalidate this rectangle, so flush draws this to screen
        invalidateRect(left, top, width, height);

        // repaint();
      }
    }
  }


  // the common method for rectEdged (for possible double operation)
  private static void rectEdged1(Color fillColour, Color frameColour,
                                 int left, int top, 
                                 int width, int height, Graphics g) {
    // draw fill rectange of given colour
    rectSolid1(fillColour, left, top, width, height, g);

    // overwrite frame of given colour
    rectOutline1(frameColour, left, top, width, height, g);
  }



  // draw a solid rectangle on given bitmap handle
  public static void rectSolid(int bitmapHandle, int colour, 
                               short left, short top,
                               short right, short bottom) {
    Log.fine("rectSolid left top", left, top);
    Log.fine("rectSolid right bottom", right, bottom);
    Log.fine("rectSolid colour", colour);

    // get the colour object
    Color c = getColorFromBGR(colour);

    // compute width and height
    int width = right-left;
    int height = bottom-top;

    // draw a filled rect
    rectSolid1(c, left, top, width, height, 
               getImage(bitmapHandle).getGraphics());

    // if screen involved, draw direct or repaint
    if (bitmapHandle == 0) {
      if (directDraw) {
        rectSolid1(c, left, top, width, height, kPanel.getGraphics());
      }
      else {
        // invalidate this rectangle, so flush draws this to screen
        invalidateRect(left, top, width, height);

        // repaint();
      }
    }
  }


  // the common method for rectSolid (for possible double operation)
  private static void rectSolid1(Color colour, int left, int top, 
                                 int width, int height, Graphics g) {
    // set the colour
    g.setColor(colour);

    // draw a solid rectangle
    g.fillRect(left, top, width, height);
  }


  // draw a outline rectangle on given bitmap handle
  public static void rectOutline(int bitmapHandle, int colour, 
                                 short left, short top,
                                 short right, short bottom) {
    Log.fine("rectOutline left top", left, top);
    Log.fine("rectOutline right bottom", right, bottom);
    Log.fine("rectOutline colour", colour);

    // get the colour object
    Color c = getColorFromBGR(colour);

    // compute width and height
    int width = right-left;
    int height = bottom-top;

    // draw a rect
    rectOutline1(c, left, top, width, height, 
                 getImage(bitmapHandle).getGraphics());

    // if screen involved, draw direct or repaint
    if (bitmapHandle == 0) {
      if (directDraw) {
        rectOutline1(c, left, top, width, height, kPanel.getGraphics());
      }
      else {
        // invalidate this rectangle, so flush draws this to screen
        invalidateRect(left, top, width, height);

        // repaint();
      }
    }
  }


  // the common method for rectOutline (for possible double operation)
  private static void rectOutline1(Color colour, int left, int top, 
                                   int width, int height, Graphics g) {
    // set the colour
    g.setColor(colour);

    // draw the rectangle outline
    g.drawRect(left, top, width, height);
  }



  // draw a poly line on given bitmap handle of given colour
  // also give rectangle that includes these lines, for invalidation
  public static void drawPolyline(int bitmapHandle, int colour, 
                                  int[]xs, int[]ys, int nPoints,
                                  Rectangle rect) {
    Log.fine("drawPolyline colour", colour);

    // compute the java colour
    Color c = getColorFromBGR(colour);

    // draw the lines
    drawPolyline1(c, xs, ys, nPoints, getImage(bitmapHandle).getGraphics());

    // if screen involved, draw direct or repaint
    if (bitmapHandle == 0) {
      if (directDraw) {
        drawPolyline1(c, xs, ys, nPoints, kPanel.getGraphics());
      }
      else {
        // invalidate the rectangle, so flush draws this to screen
        invalidateRect(rect.x, rect.y, rect.width, rect.height);

        // repaint();
      }
    }
  }


  // the common method for drawPolyline (for possible double operation)
  private static void drawPolyline1(Color colour, int[]xs, int[]ys, int nPoints,
                                    Graphics g) {
    // set the colour
    g.setColor(colour);

    // draw a poly line
    g.drawPolyline(xs, ys, nPoints);
  }



  // support routine to convert y of KINT to y of java for drawString
  // also used by Printing
  public static int kintTextY2Java(int y, Font f, Graphics g) {
    int result = y + g.getFontMetrics(f).getAscent();
    Log.fine("kintTextY2Java kint_y java_y", y, result);
    return result;
  }



  // draw a string at given position on given bitmap handle
  public static void drawString(int bitmapHandle, String s, short x, short y) {
    Log.fine("drawString string", s);
    Log.fine("drawString location", x, y);

    // unfortunately we need the graphics here, so we can compute font metrics
    // and recalc newy and box for opaque
    // we could do in drawString1 but that would repeat it.  
    Graphics g = getImage(bitmapHandle).getGraphics();

    // it seems KInt co-ords are top-left of text, where as Java are the
    // base-line of the left most character, so move y down the ascent
    int newy = kintTextY2Java(y, currentFont, g);

    // if back mode is OPAQUE we set the back colour to drawString1 so it
    // can draw a filled rect, we compute the width/height here
    boolean drawBack = false;
    int w = 0;
    int h = 0;
    if (currentBackMode == TextState.BACKMODE_OPAQUE) {
      drawBack = true;
      FontMetrics fm = g.getFontMetrics(currentFont);
      w = fm.stringWidth(s);
      h = fm.getHeight();
      Log.fine("opaque mode w h", w, h);
    }

    // draw the string
    drawString1(currentTextColour, currentBackColour, s, x, y, newy, w, h,
                currentFont, drawBack, g);

    // if screen involved, draw direct or repaint
    if (bitmapHandle == 0) {
      if (directDraw) {
        drawString1(currentTextColour, currentBackColour, s, x, y, newy, w, h, 
                    currentFont, drawBack, kPanel.getGraphics());
      }
      else {
        // No repaint or invalidate rect here, as KINT does not do this
        // Uses of KINT make sure text is only drawn onto an already invalid
        // area
        // repaint();
      }
    }
  }


  // the common method for drawString (for possible double operation)
  private static void drawString1(Color textColour, Color backColour, String s, 
                                  int x, int y, int newy, int w, int h,
                                  Font font, boolean drawBack, Graphics g) {
    // draw the background if required
    if (drawBack) {
      g.setColor(backColour);
      g.fillRect(x, y, w, h);
    }

    // set the font
    g.setFont(font);

    // set the colour for the text
    g.setColor(textColour);

    // draw the string
    g.drawString(s, x, newy);
  }



  // cause a repaint, only if not in direct draw mode
  public static void repaint() {
    if (!directDraw) {
      if (kPanel != null) {
        Log.info("UI.repaint calling kPanel.repaint");

        // do nothing for now - reconsidering buffered drawing
        // kPanel.repaint();
      }
    }
  }



  // store the bitmap, returning its handle
  private static int storeBitmap(BMB bmb) {
    return Utils.addObjectToHandles(bmb, bmbs, false, "BMBs");
  }


  // get the BMB for the given bitmap handle
  // 0 represents the game's display area
  public static BMB getBMB(int bmbHandle) {
    BMB bmb = (BMB)Utils.getObjectByHandle(bmbHandle, bmbs, false, "BMBs");
    
    // we must always have a result
    if (bmb == null) {
      Log.errorExit("Null BMB for bmb handle "+bmbHandle+", perhaps that handle has been deleted.");
    }
               
    return bmb;
  }



  //*************************************************
  // Printer stuff
  //*************************************************

  public static PrintJob getPrintJob() {
    Log.fine("getPrintJob");
    return jFrame.getToolkit().getPrintJob(jFrame, "KINT", null);
  }
  




  //*************************************************
  // Debug stuff
  //*************************************************

  // show the string in a modal dialog
  public static void showMessageDialog(String msg) {
    Log.fine("showMessageDialog", msg);
    JOptionPane.showMessageDialog(jFrame, msg, "KInt Debug Message",
                                  JOptionPane.WARNING_MESSAGE);
  }
      



  //*************************************************
  // FONTS
  //*************************************************


  // return the font index for the font that fits the given information
  // work's relative to screen's (window's) graphics object
  public static short fitFont(String s, int w, int h, int style) {
    Log.fine("fitFont for", s);
    Log.fine("location", w, h);
    Log.fine("style", style);

    // decide if variable or fixed pitch font
    String fontName = "Arial";
    if ((style & 0x01) != 0) {
      fontName = "Courier New";
    }
    
    // decide if bold and/or italic
    int fontStyle = Font.PLAIN;
    if ((style & 0x02) != 0) {
      if ((style & 0x04) != 0) {
        fontStyle = Font.BOLD | Font.ITALIC;
      }
      else {
        fontStyle = Font.BOLD;
      }
    }
    else if ((style & 0x04) != 0) {
      fontStyle = Font.ITALIC;
    }

    // if the string is empty (or null) then just choose size 8
    if (s == null || s.length() == 0) {
      return addFont(new Font(fontName, fontStyle, 12));
    }

    // will now use a similar algorithm to Kaser's
    // start at size 10, adding 2 until too big, then reduce by 2 until fits
    // in or size is 4
    // we could do more with 2D graphics here, rather than using straight AWT
    // methods
    // works based on screen's graphics 
    int size = 8;
    Font font = null;
    Graphics g = getImage(0).getGraphics();

    // look for the size that is too big
    do {
      size += 2;
      font = new Font(fontName, fontStyle, size);
      FontMetrics fm = g.getFontMetrics(font);
      if (fm.stringWidth(s)>= w || fm.getHeight()>=h) break;
    }
    while (true);
    Log.fine("fitFont, too big font size is ", size);

    // now look smaller until fits in, or size is 4
    do {
      size -= 2;
      font = new Font(fontName, fontStyle, size);
      FontMetrics fm = g.getFontMetrics(font);
      if (size<=4 || (fm.stringWidth(s)< w && fm.getHeight()<h)) break;
    }
    while (true);

    Log.fine("fitFont, selected font size is ", size);
    return addFont(font);
  }


  // get the font size - not needed
  // public static short getFontSize(short fontIndex) {
  //  return (short)getFont(fontIndex).getSize();
  // }


  // get the font height
  // all measurements done from screen graphics
  public static int getFontHeight(short fontIndex) {
    return getImage(0).getGraphics().getFontMetrics(getFont(fontIndex)).getHeight();
  }


  // get the font ascent, i.e. height of non-descending portion of font
  // all measurements done from screen graphics
  public static int getFontAscent(short fontIndex) {
    return getImage(0).getGraphics().getFontMetrics(getFont(fontIndex)).getAscent();
  }


  // get the text size information
  // work on the screen's graphics
  public static int getTextSize(String s, Dimension d) {
    Log.fine("getTextSize for string", s);

    // work on screen's image
    Graphics g = getImage(0).getGraphics();

    // work with current font in this graphics context
    return getTextSize(s, d, currentFont, g);
  }


  // common method so printer can use same functionality
  public static int getTextSize(String s, Dimension d, Font f, Graphics g) {
    FontMetrics fm = g.getFontMetrics(f);

    // width is based on the string
    int width = fm.stringWidth(s);
    Log.fine("getTextSize width", width);

    // height is from the font
    int height = fm.getHeight();
    Log.fine("getTextSize height", height);

    // set return dimension
    d.setSize(width, height);
    
    // ascent is from the font
    int ascent = fm.getAscent();
    Log.fine("getTextSize ascent", ascent);

    return ascent;
  }

  
  // set the current font, given the font
  public static void setCurrentFont(Font f) {
    currentFont = f;
    // Log.fine("Current font set to "+f);
  }


  // release the font, given the handle
  public static void releaseFont(short fontHandle) {
    // just set it to null, and let the garbage collector do its work!
    Utils.removeObjectByHandle(fontHandle, fonts, false, "fonts");
  }


  // add font to fonts vector and return handle
  private static short addFont(Font f) {
    int fontHandle = Utils.addObjectToHandles(f, fonts, false, "fonts");
    return (short)fontHandle;
  }


  // get the font at the handle - must be a font here
  public static Font getFont(short fontHandle) {
    Font f = (Font)Utils.getObjectByHandle(fontHandle, fonts, false, "fonts");
    if (f == null) {
      Log.errorExit("No font matching handle", fontHandle);
    }
    return f;
  }


  //***************************************************
  // Colour methods
  //***************************************************

  // set the current text colour
  public static void setCurrentTextColour(Color c) {
    // Log.fine("setCurrentTextColour "+c);
    currentTextColour = c;
  }


  // set the current text back colour
  public static void setCurrentBackColour(Color c) {
    // Log.fine("setCurrentTextColour "+c);
    currentBackColour = c;
  }

  // set the current text back mode (no check for valid arg)
  public static void setCurrentBackMode(int m) {
    Log.fine("setCurrentBackMode ", m);
    currentBackMode = m;
  }


  // support in method for rgb and bgr conversion

  // swapped r and b (dropping bits 24-31)
  private static int swapRedAndBlue(int rgb) {
    int red = rgb & 0xFF0000;
    int green = rgb & 0xFF00;
    int blue = rgb & 0xFF;
    return (blue << 16) | green | (red >> 16);
  }

  // get colour given BGR value
  public static Color getColorFromBGR(int bgr) {
    return new Color(swapRedAndBlue(bgr));
  }



  //******************************************************
  // Text States 
  //******************************************************


  // set the current text style to given args
  public static void setTextStyle(int textColor, int backColor, 
                                  short fontHandle) {
    Log.fine("setTextStyle textColor backColor", textColor, backColor);

    // set the text colour
    setCurrentTextColour(getColorFromBGR(textColor));

    // a backColor of -1 means transparent
    if (backColor == -1) {
      // transparent has no color
      setCurrentBackMode(TextState.BACKMODE_TRANSPARENT);
    }
    else {
      setCurrentBackMode(TextState.BACKMODE_OPAQUE);
      setCurrentBackColour(getColorFromBGR(backColor));
    }

    // set the current font handle
    setCurrentFont(getFont(fontHandle));
  }


  // save the text state 
  public static void saveTextState() {
    textStates.insertElementAt(new TextState(currentTextColour, 
                                             currentBackColour, 
                                             currentBackMode, 
                                             currentFont), 0);
  }


  // restore (then discard) the top text state 
  public static void restoreTextState() {
    // must be one, otherwise no previous SYS_BEGTEXT
    if (textStates.size() == 0) {
      Log.errorExit("No Text State to return, missing SYS_BEGTEXT");
    }

    // get and remove top text state
    TextState ts = (TextState)textStates.elementAt(0);
    textStates.removeElementAt(0);

    // restore 
    setCurrentTextColour(ts.getTextColour());
    setCurrentBackColour(ts.getBackColour());
    setCurrentBackMode(ts.getBackMode());
    setCurrentFont(ts.getFont());
  }



  //*************************************************
  // Timer functions
  //*************************************************

  // create the timer if not already in existence
  public static void createTimer(int period) {
    if (timer == null) {
      // System.out.println("Creating timer with period "+period);
      Log.fine("Creating timer with period", period);
      timer = new javax.swing.Timer(period, new KTimerListener());
      timer.start();
    }
    else {
      Log.severe("Timer already exists, cannot create another");
    }
  }

  // delete the timer if it exists
  public static void deleteTimer() {
    if (timer != null) {
      Log.fine("Deleting timer");
      timer.stop();
      timer = null;
    }
  }


  //*************************************************
  // Event Queue Methods
  //*************************************************

  // return true if there are no events waiting in the AWT event queue
  public static boolean isAWTEventQueueEmpty() {
    if (awtEventQueue == null) {
      awtEventQueue = Toolkit.getDefaultToolkit().getSystemEventQueue();
    }

    AWTEvent e = awtEventQueue.peekEvent();
    boolean result = (e == null);
    Log.fine("isAWTEventQueueEmpty returning", result);
    return result;
  }

  
  // wait an AWT Event to happen (so that we can then continue to look
  // for an idle moment).
  //
  // We first look to see if event queue has something on it, if so we
  // can return.
  // We then keep sleeping.  On wake up we check the time of the last event
  // until that last event time is greater than when we entered this method.
  // Time of last event is set in the event displatch thread by our event
  // listeners.
  //
  // We could have (but do not) a listener that listens for all (relevant)
  // AWT events and for that listener to set the last time.
  //
  // I cannot use the EventQueue's getMostRecentEventTime() as that only 
  // works when called from the event dispatch thread, and I don't want this
  // thread to send event code to that thread, else it itself sets the last
  // event time.
  //
  // Another options is for this thread to go into a wait() and for our
  // event processing code to notify() it.
  //
  // Note: I tried the simple "sleep and then see if event queue has something
  // in it" approach - but this did not work, as events were processed off
  // the queue very quickly (and I did not want a low sleep time as we would
  // then become very busy).
  public static void waitForAWTEventToHappen() {
    Log.fine("Waiting for AWTEvent to happen: IN");

    // remember now, so we can see if any events happened after
    long timeIn = System.currentTimeMillis();

    if (!isAWTEventQueueEmpty()) {
      Log.fine("AWT EventQueue was immediately non-empty");
      Log.fine("Waiting for AWTEvent to happen: OUT");
      return;
    }

    while(mostRecentEventTime < timeIn) {
      // sleep and check
      try {
        Log.fine("sleeping for 1 second");
        Thread.sleep(1000);
      }
      catch (InterruptedException e) {
        Log.errorExit("Should not have been interrupted in waitForAWTEventToHappen");
      }
    }

    Log.fine("Waiting for AWTEvent to happen: OUT");
  }


  // wait for the AWT EventQueue to become empty
  // For now take a simple "sleep and keep checking" approach
  // Of course this could mean we miss the AWT EventQueue becoming empty
  // i.e. it could have gone empty but has now got some more event.
  // This is okay as we are only interested in IDLE time
  public static void waitForEmptyAWTEventQueue() {
    Log.fine("Waiting for empty AWT EventQueue: IN");

    if (isAWTEventQueueEmpty()) {
      Log.fine("AWT EventQueue was immediately empty");
      Log.fine("Waiting for empty AWT EventQueue: OUT");
      return;
    }

    while(!isAWTEventQueueEmpty()) {
      // sleep and check
      try {
        Log.fine("sleeping for 1 second");
        Thread.sleep(1000);
      }
      catch (InterruptedException e) {
        Log.errorExit("Should not have been interrupted in waitForEmptyAWTEventQueue");
      }
    }

    Log.fine("Waiting for empty AWT EventQueue: OUT");
  }


  // set the most recent event time to now - called when event noted
  public static void setMostRecentEventTime() {
    mostRecentEventTime = System.currentTimeMillis();
  }


  // *************** for some simple performance **************8
  public static long perfLoadBitmap() {
    return loadBitmapTime;
  }

  public static long perfLoadMBitmap() {
    return loadMBitmapTime;
  }

  //***************** for testing ***************
  //  public static void seeTraversalKeys() {
  //    System.out.println("getFocusTraversalKeysEnabled()="+jFrame.getFocusTraversalKeysEnabled());
  //    System.out.println("FORWARD_TRAVERSAL_KEYS: "+jFrame.getFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS));
  //    System.out.println("BACKWARD_TRAVERSAL_KEYS: "+jFrame.getFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS));
  //    System.out.println("UP_CYCLE_TRAVERSAL_KEYS: "+jFrame.getFocusTraversalKeys(KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS));
  //  }


  // Return our CVS id
  public static String getCVSID() {
    return "$Id: UI.java,v 1.6 2002/11/28 19:32:28 Bennett Stephen Exp $";
  }


}


/*
 * $Log: UI.java,v $
 * Revision 1.6  2002/11/28 19:32:28  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.5  2002/10/12 14:21:29  Bennett Stephen
 * corrected font metrics
 *
 * Revision 1.4  2002/10/06 15:24:02  Bennett Stephen
 * safe version, still being worked on
 *
 * Revision 1.3  2002/09/08 19:39:08  Bennett Stephen
 * mainly addition of bitmap and graphic functions
 *
 * Revision 1.2  2002/08/10 08:24:49  Bennett Stephen
 * turn off focus enable so handle own TAB char
 *
 * Revision 1.1  2002/08/07 19:22:19  Bennett Stephen
 * Initial Version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
