// A Java version for the Kaser Interpreter - load .BMP to ImageProducer
// For some of my own fun!
// Stephen M. Bennett
// 15Oct02 Started


package stephen.bennett.kint;

import java.io.*;
import java.awt.image.*;


public class WindowsBMP {


  // a static method to do the work, similar to ImageIO.read(File f)
  // no transparency required
  public static BufferedImage read(File f) throws IOException {
    // just call the main work method in the right way
    return read(f, false, -1, -1);
  }


  // read the image, and convert all pixels with same colour as named pixel
  // to transparent
  public static BufferedImage read(File f, int transX, int transY) throws IOException {
    // just call the main work method in the right way
    return read(f, true, transX, transY);
  }


  // Here we do the read, doing transparent conversion if requested
  private static BufferedImage read(File f, boolean trans, 
                                    int transX, int transY) throws IOException {
    FileInputStream fis = new FileInputStream(f);

    // remember where we are in the file reads
    int fOffset = 0;

    // first comes BM
    int c1 = fis.read();
    int c2 = fis.read();
    if (c1 != 'B' || c2 != 'M') {
      throw new IOException("File not in BMP format");
    }
    fOffset+=2;

    // file size next
    int bmpFileSize = read32(fis);
    fOffset+=4;

    Log.fine("Windows BMP file size", bmpFileSize);

    // ignore the two reserve 16 bits
    read32(fis);
    fOffset+=4;

    // off set for the bitmaps next
    int bmpStartOffset = read32(fis);
    fOffset+=4;

    // now into header info

    // the size of the header
    int size = read32(fis);
    fOffset+=4;
    if (size != 40) {
      // I only expect this header size
      throw new IOException("Cannot cope with BMP header of size "+size+" (decimal)");
    }

    // the width and height
    int width = read32(fis);
    fOffset+=4;
    int height = read32(fis);
    fOffset+=4;

    Log.fine("Windows BMP width height", width, height);

    // the number of planes
    int planes = read16(fis);
    fOffset+=2;
    if (planes != 1) {
      // I can only cope with 1 plane
      throw new IOException("Cannot cope with BMP with "+planes+" planes");
    }

    // get the number of bits 
    int bitCount = read16(fis);
    fOffset+=2;
    if (bitCount != 1 && bitCount != 8 && bitCount != 24) {
      // I've only implemented 1, 8 and 24 at the moment
      throw new IOException("Cannot cope with BMP bbp "+bitCount+" (decimal)");
    }

    Log.fine("Windows BMP bpp", bitCount);

    // scan line is padded with zeroes to be a multiple of four bytes
    int scanLineSize = ((width*bitCount+31)/32)*4;
    Log.fine("Windows BMP scan line size", scanLineSize);

    // the compression algorithm
    int compression = read32(fis);
    fOffset+=4;
    if (compression != 0) {
      // I don't have a clue about compressions
      throw new IOException("Cannot cope with BMP compression "+compression+" (decimal)");
    }

    // the size of the bitmap image
    int bmpSize = read32(fis);
    fOffset+=4;
    Log.fine("Windows BMP bitmap size", bmpSize);

    // if its 0 then need to work it out ourself
    if (bmpSize == 0) {
      bmpSize = scanLineSize * height;
      Log.fine("Windows BMP new bitmap size", bmpSize);
    }

    // read the resolutions
    int hRes = read32(fis);
    fOffset+=4;
    int vRes = read32(fis);
    fOffset+=4;

    // read the number of colours used (0 means work out yourself)
    int colorsUsed = read32(fis);
    fOffset+=4;
    if (colorsUsed == 0) {
      // if less then 16 bits then there will be colours, otherwise no palette
      if (bitCount < 16) {
        colorsUsed = 1 << bitCount;
      }
    }

    // read the number of important colours (0 means all)
    int colorsImportant = read32(fis);
    fOffset+=4;

    // read the palette, if there are colours
    int[] palette = null;
    if (colorsUsed > 0) {
      // claim the space
      palette = new int[colorsUsed];

      // read the colours in the palette
      int b, g, r, c;
      for (int i = 0; i < colorsUsed; i++) {
        b = fis.read();
        g = fis.read();
        r = fis.read();
        // ignore the 4th bit
        fis.read();
        fOffset+=4;

        // compute this colour
        palette[i] = 0xff000000 | (r<<16) | (g<<8) | b;
      }
    }

    // now read the pixels, first need to get to where we want to be
    // need to skip up to start of bitmap
    int skip = bmpStartOffset - fOffset;
    if (skip > 0) {
      fis.skip(skip);
    }

    // read bitmap bytes all in, in one big suck (for effeciency)
    byte[] allBytes = new byte[bmpSize];
    int n = fis.read(allBytes, 0, bmpSize);
    if (n < bmpSize) {
      throw new IOException("Could not read all bitmap in one go");
    }

    // get the transparent colour, if required
    // transValue may be index into palette or actual RGB value (with no alpha)
    int transValue = -1;
    // transMatch is for 24 bit being the colour with the alpha channel
    // for effeciency
    int transMatch = -1;
    if (trans) {
      if (bitCount == 1) {
        // calculate the offset where the transparent colour index is
        // remember first row is height-1 row
        int transOffset = ((height-1-transY)*scanLineSize)+(transX/8);
        //System.out.println("transOffset="+transOffset);

        // get the byte here
        int b = allBytes[transOffset++] & 0xFF;

        // we only want THE bit of this, being the index into the palette
        int bOffset = transX%8;
        for(int i=0; i<bOffset; i++) {
          b <<=1;
        }
        
        int colourIndex = (b & 0x80) != 0 ? 1 : 0;

        // turn the alpha off for that colour in the palette
        palette[colourIndex] = palette[colourIndex] & 0xFFFFFF;
      }
      else if (bitCount == 8) {
        // calculate the offset where the transparent colour index is
        // remember first row is height-1 row
        int transOffset = ((height-1-transY)*scanLineSize)+transX;
        //System.out.println("transOffset="+transOffset);

        // the byte here is the colour index
        int colourIndex = allBytes[transOffset++] & 0xFF;

        // turn the alpha off for that colour in the palette
        palette[colourIndex] = palette[colourIndex] & 0xFFFFFF;
      }
      else if (bitCount == 24) {
        // calculate the offset where the transparent colour starts
        // remember first row is height-1 row
        int transOffset = ((height-1-transY)*scanLineSize)+(transX*3);
        //System.out.println("transOffset="+transOffset);

        // get the colour here
        int b = allBytes[transOffset++] & 0xFF;
        int g = allBytes[transOffset++] & 0xFF;
        int r = allBytes[transOffset++] & 0xFF;
        transValue = (r<<16) | (g<<8) | b;
        transMatch = 0xff000000 | transValue;
        //System.out.println("transparent colour="+Integer.toHexString(transValue));
      }
      else {
        throw new IOException("Cannot cope with BMP bbp "+bitCount+" (decimal)");
      }
    }

    // we need to remember our offset into these bytes
    int allBytesOffset = 0;

    // we build up the BufferedImage as we read each pixel
    BufferedImage result = new BufferedImage(width, height, 
                                             BufferedImage.TYPE_INT_ARGB);

    // here are the algorithms to read the file based bbp
    if (bitCount == 1) {
      // each bit is 0 or 1 colour index

      // lines are multiple of 32 bits
      int padding = ((width+31)/32)*32;

      // do the read
      int x, y, c = 0, bit;
      for(y=height-1; y>=0; y--) {
        for(x=0, bit=0; x<padding; x++, bit++) {
          // read the next byte, if we have done the last bit
          if ((bit & 0x07) == 0) {
            c = allBytes[allBytesOffset++];

            // check its not EOF - not now
            // if (c < 0) {
            //   throw new EOFException();
            // }
            
            // back to the first bit
            bit = 0;
          }

          // if not reading end of line padding, remember this
          if (x < width) {
            // set this pixel
            int colourIndex = (c & 0x80) != 0 ? 1 : 0;

            // don't care about transparency here because the palette holds
            // the transparent or non-transparent colour already
            result.setRGB(x, y, palette[colourIndex]);
            c <<=1;
          }
        }
      }
    }
    else if (bitCount == 8) {
      // byte values of indexes into the palette

      // lines are multiple of 32 bits
      int lineSize = ((width+3)/4)*4;

      // do the read
      int x, y, c;
      for(y=height-1; y>=0; y--) {
        for(x=0; x<lineSize; x++) {
          // read the next byte
          c = allBytes[allBytesOffset++] & 0xFF;

          // check its not EOF - no will get index exception now
          // if (c < 0) {
          //   throw new EOFException();
          // }

          // if not reading end of line padding, remember this
          if (x < width) {
            // don't care about transparency here because the palette holds
            // the transparent or non-transparent colour already

            // set this pixel
            result.setRGB(x, y, palette[c]);
          }
        }
      }
    }
    else if (bitCount == 24) {
      // RGB is 4 bytes, explicit (i.e. no use of palette)

      // lines are padded so
      int padding = (4-((width*3)%4))&0x03;

      // do the read
      int x, y, b, g, r, c;
      for(y=height-1; y>=0; y--) {
        for(x=0; x<width; x++) {
          // read colours
          b = allBytes[allBytesOffset++] & 0xFF;   
          g = allBytes[allBytesOffset++] & 0xFF;   
          r = allBytes[allBytesOffset++] & 0xFF;   

          // check its not EOF
          // if (b < 0 || g < 0 || r < 0) {
          //   throw new EOFException();
          // }

          // set this pixel
          c = 0xff000000 | (r<<16) | (g<<8)  | b;

          // if this is the transparent colour then don't want alpha channel
          if (trans) {
            if (c == transMatch) {
              // drop the alpha channel
              c = transValue;  
            }
          }

          result.setRGB(x, y, c);
        }
          
        // skip the padding
        for(x=0; x<padding; x++) {
          allBytesOffset++;   
        }
      }
    }
    else {
      throw new IOException("Cannot cope with BMP bbp "+bitCount+" (decimal)");
    }

    return result;
  }



  // read 32 bits from given input stream
  private static int read32(InputStream is) throws IOException {
    int c1 = is.read();
    int c2 = is.read();
    int c3 = is.read();
    int c4 = is.read();
    return (c4<<24) | (c3<<16) | (c2<<8) | c1;
  }


  // read 16 bites from given input stream
  private static int read16(InputStream is) throws IOException {
    int c1 = is.read();
    int c2 = is.read();
    return (c2<<8) | c1;
  }



  // Return our CVS id
  public static String getCVSID() {
    return "$Id: WindowsBMP.java,v 1.1 2002/10/19 16:08:33 Bennett Stephen Exp $";
  }


}


/*
 * $Log: WindowsBMP.java,v $
 * Revision 1.1  2002/10/19 16:08:33  Bennett Stephen
 * Initial version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
